import os
import random
from dataclasses import dataclass

from topk_sdk import Client


@dataclass
class ProjectContext:
    client: Client
    scope_prefix: str

    def scope(self, name: str):
        return f"{self.scope_prefix}-{name}"


def new_project_context():
    TOPK_API_KEY = os.environ["TOPK_API_KEY"].splitlines()[0].strip()
    TOPK_HOST = os.environ.get("TOPK_HOST", "topk.io")
    TOPK_REGION = os.environ.get("TOPK_REGION", "elastica")

    client = Client(
        api_key=TOPK_API_KEY,
        region=TOPK_REGION,
        host=TOPK_HOST,
    )

    return ProjectContext(
        scope_prefix="%s" % str(random.random()).replace(".", ""),
        client=client,
    )



import pytest
from topk_sdk.query import field, fn, match, select
from topk_sdk.schema import keyword_index, text, f32_vector, u8_vector, binary_vector, vector_index

from . import ProjectContext


def test_upsert(ctx: ProjectContext):
    ctx.client.collections().create(ctx.scope("books"), schema={})

    lsn = ctx.client.collection(ctx.scope("books")).upsert(
        [
            {"_id": "one", "name": "one"},
            {"_id": "two", "name": "two", "extra": 1},
        ],
    )
    assert lsn == 1

    docs = ctx.client.collection(ctx.scope("books")).query(
        select("name").filter(field("name").eq("two")),
        lsn=lsn,
    )

    assert docs == [{"_id": "two", "name": "two"}]


def test_keyword_search(ctx: ProjectContext):
    ctx.client.collections().create(
        ctx.scope("books"),
        schema={
            "title": text().required().index(keyword_index()),
        },
    )

    lsn = ctx.client.collection(ctx.scope("books")).upsert(
        [
            {"_id": "doc1", "title": "red purple green"},
            {"_id": "doc2", "title": "yellow purple pink"},
            {"_id": "doc3", "title": "orange red blue"},
            {"_id": "doc4", "title": "green yellow purple"},
            {"_id": "doc5", "title": "pink orange red"},
            {"_id": "doc6", "title": "black green yellow"},
            {"_id": "doc7", "title": "purple pink orange"},
            {"_id": "doc8", "title": "red yello green"},
            {"_id": "doc9", "title": "yellow purple pink"},
            {"_id": "doc10", "title": "orange red blue"},
        ],
    )
    assert lsn == 1

    docs = ctx.client.collection(ctx.scope("books")).query(
        select(
            text_score=fn.keyword_score(),
        )
        .filter(
            match("title", token="blue", weight=10.0)
            | match("title", token="red", weight=50.0),
        )
        .top_k(field("text_score"), k=5),
        lsn=lsn,
    )

    assert {d["_id"] for d in docs} == {"doc1", "doc10", "doc3", "doc5", "doc8"}


def test_vector_search_f32(ctx: ProjectContext):
    ctx.client.collections().create(
        ctx.scope("books"),
        schema={
            "f32_embedding": f32_vector(3).required().index(vector_index(metric="euclidean")),
        },
    )

    lsn = ctx.client.collection(ctx.scope("books")).upsert(
        [
            {"_id": "doc1", "f32_embedding": [1.0, 2.0, 3.0]},
            {"_id": "doc2", "f32_embedding": [4.0, 5.0, 6.0]},
            {"_id": "doc3", "f32_embedding": [7.0, 8.0, 9.0]},
        ],
    )
    assert lsn == 1

    docs = ctx.client.collection(ctx.scope("books")).query(
        select(
            vector_distance=fn.vector_distance("f32_embedding", [7.0, 8.0, 9.0]),
        ).top_k(field("vector_distance"), k=2, asc=True),
        lsn=lsn,
    )
    docs.sort(key=lambda d: d["vector_distance"])

    assert [d["_id"] for d in docs] == ["doc3", "doc2"]

def test_vector_search_u8(ctx: ProjectContext):
    ctx.client.collections().create(
        ctx.scope("books"),
        schema={
            "u8_embedding": u8_vector(3).required().index(vector_index(metric="euclidean")),
        },
    )

    lsn = ctx.client.collection(ctx.scope("books")).upsert(
        [
            {"_id": "doc1", "u8_embedding": [1, 2, 3]},
            {"_id": "doc2", "u8_embedding": [4, 5, 6]},
            {"_id": "doc3", "u8_embedding": [7, 8, 9]},
        ],
    )
    assert lsn == 1

    docs = ctx.client.collection(ctx.scope("books")).query(
        select(
            vector_distance=fn.vector_distance("u8_embedding", [7, 8, 9]),
        ).top_k(field("vector_distance"), k=2, asc=True),
        lsn=lsn,
    )
    docs.sort(key=lambda d: d["vector_distance"])

    assert [d["_id"] for d in docs] == ["doc3", "doc2"]

def test_vector_search_binary(ctx: ProjectContext):
    ctx.client.collections().create(
        ctx.scope("books"),
        schema={
            "binary_embedding": binary_vector(3).required().index(vector_index(metric="hamming")),
        },
    )

    lsn = ctx.client.collection(ctx.scope("books")).upsert(
        [
            {"_id": "doc1", "binary_embedding": [0, 0, 1]},
            {"_id": "doc2", "binary_embedding": [0, 1, 1]},
            {"_id": "doc3", "binary_embedding": [1, 1, 1]},
        ],
    )
    assert lsn == 1

    docs = ctx.client.collection(ctx.scope("books")).query(
        select(
            vector_distance=fn.vector_distance("binary_embedding", [1, 1, 1]),
        ).top_k(field("vector_distance"), k=2, asc=True),
        lsn=lsn,
    )
    docs.sort(key=lambda d: d["vector_distance"])

    assert [d["_id"] for d in docs] == ["doc3", "doc2"]


def test_delete(ctx: ProjectContext):
    ctx.client.collections().create(ctx.scope("books"), schema={})

    lsn = ctx.client.collection(ctx.scope("books")).upsert(
        [
            {"_id": "doc1", "name": "one"},
        ],
    )
    assert lsn == 1

    lsn = ctx.client.collection(ctx.scope("books")).delete(["doc1"])
    assert lsn == 2

    docs = ctx.client.collection(ctx.scope("books")).query(
        select("name").filter(field("name").eq("one")),
        lsn=lsn,
    )
    assert docs == []


@pytest.fixture
def ctx():
    from . import new_project_context

    return new_project_context()



import pytest
from topk_sdk import Collection, error
from topk_sdk.schema import (
    binary_vector,
    bool,
    bytes,
    f32_vector,
    float,
    int,
    keyword_index,
    text,
    u8_vector,
    vector_index,
)

from . import ProjectContext


def test_create_collection(ctx: ProjectContext):
    schema = {
        "title": text().required().index(keyword_index()),
        "title_embedding": f32_vector(1536)
        .required()
        .index(vector_index(metric="euclidean")),
        "published_year": int().required(),
    }

    collection = ctx.client.collections().create(
        ctx.scope("books"),
        schema=schema,
    )

    assert collection.name == ctx.scope("books")
    assert collection.schema == schema


def test_create_collection_all_data_types(ctx: ProjectContext):
    schema = {
        "text": text(),
        "int": int(),
        "float": float(),
        "bool": bool(),
        "vector": f32_vector(1536),
        "float_vector": f32_vector(1536),
        "byte_vector": u8_vector(1536),
        "binary_vector": binary_vector(1536),
        "bytes": bytes(),
    }

    collection = ctx.client.collections().create(
        ctx.scope("books"),
        schema=schema,
    )

    assert collection.name == ctx.scope("books")
    assert collection.schema == schema


def test_incorrect_schema(ctx: ProjectContext):
    with pytest.raises(error.SchemaValidationError):
        ctx.client.collections().create(
            ctx.scope("books"),
            schema={"name": text().index(vector_index(metric="cosine"))},
        )


def test_list_collections(ctx: ProjectContext):
    # Note: All python tests run within the same project,
    # so list of collections is shared across tests.

    a = ctx.client.collections().create(ctx.scope("books"), schema={})
    assert a in ctx.client.collections().list()

    b = ctx.client.collections().create(ctx.scope("books2"), schema={})
    assert a in ctx.client.collections().list()
    assert b in ctx.client.collections().list()

    c = ctx.client.collections().create(ctx.scope("books3"), schema={})
    assert a in ctx.client.collections().list()
    assert b in ctx.client.collections().list()
    assert c in ctx.client.collections().list()


def test_delete_collection(ctx: ProjectContext):
    assert ctx.scope("books") not in [c.name for c in ctx.client.collections().list()]

    ctx.client.collections().create(ctx.scope("books"), schema={})

    assert ctx.scope("books") in [c.name for c in ctx.client.collections().list()]

    ctx.client.collections().delete(ctx.scope("books"))

    assert ctx.scope("books") not in [c.name for c in ctx.client.collections().list()]


def test_delete_non_existent_collection(ctx: ProjectContext):
    try:
        ctx.client.collections().delete(ctx.scope("books"))
    except error.CollectionNotFoundError:
        pass
    else:
        assert False, "CollectionNotFoundError not raised"


###


@pytest.fixture
def ctx():
    from . import new_project_context

    return new_project_context()



import pytest
from topk_sdk.query import field, literal


def test_query_expr_with_flexible_expr():
    assert field("a") + 1 == field("a") + literal(1)
    assert 1 + field("a") == literal(1) + field("a")

    assert field("a") - 1 == field("a") - literal(1)
    assert 1 - field("a") == literal(1) - field("a")

    assert field("a") * 1 == field("a") * literal(1)
    assert 1 * field("a") == literal(1) * field("a")

    assert field("a") / 1 == field("a") / literal(1)
    assert 1 / field("a") == literal(1) / field("a")

    assert field("a") & True == field("a") & literal(True)
    assert True & field("a") == literal(True) & field("a")

    assert field("a") | False == field("a") | literal(False)
    assert False | field("a") == literal(False) | field("a")


def test_query_expr_eq():
    assert literal("a") + literal("b") == literal("a") + literal("b")
    assert literal("a") != literal("b")
    assert field("a") == field("a")
    assert field("a") != field("b")
    assert field("a") != literal("a")
    assert literal("a") != field("a")


def test_query_literal():
    field("foo").eq(literal(1))
    field("foo").eq(1)

    field("foo").ne(literal(1))
    field("foo").ne(1)


def test_invalid():
    with pytest.raises(TypeError):
        literal(1) + "string"

    with pytest.raises(TypeError):
        field("a") & 1

    with pytest.raises(TypeError):
        field("a") | 1

    with pytest.raises(TypeError):
        field("a") + None

    with pytest.raises(TypeError):
        field("a") + "string"

    with pytest.raises(TypeError):
        field("a") + [1, 2, 3]

    with pytest.raises(TypeError):
        field("a") + {"a": 1}
